use std::sync::Arc;
use vulkano::buffer::{Buffer, BufferContents, BufferCreateInfo, BufferUsage, Subbuffer};
use vulkano::command_buffer::{AutoCommandBufferBuilder, PrimaryAutoCommandBuffer};
use vulkano::device::Device;
use vulkano::memory::allocator::{AllocationCreateInfo, MemoryTypeFilter, StandardMemoryAllocator};
use vulkano::pipeline::graphics::vertex_input::{Vertex, VertexDefinition};
use vulkano::pipeline::{DynamicState, GraphicsPipeline, PipelineLayout, PipelineShaderStageCreateInfo};
use vulkano::pipeline::graphics::color_blend::{ColorBlendAttachmentState, ColorBlendState};
use vulkano::pipeline::graphics::GraphicsPipelineCreateInfo;
use vulkano::pipeline::graphics::input_assembly::InputAssemblyState;
use vulkano::pipeline::graphics::multisample::MultisampleState;
use vulkano::pipeline::graphics::rasterization::RasterizationState;
use vulkano::pipeline::graphics::subpass::PipelineRenderingCreateInfo;
use vulkano::pipeline::graphics::viewport::ViewportState;
use vulkano::pipeline::layout::PipelineDescriptorSetLayoutCreateInfo;
use vulkano::swapchain::Swapchain;

// We use `#[repr(C)]` here to force rustc to use a defined layout for our data, as the default
// representation has *no guarantees*.
#[derive(BufferContents, Vertex)]
#[repr(C)]
struct MyVertex {
    #[format(R32G32_SFLOAT)]
    position: [f32; 2],

    #[format(R32G32B32_SFLOAT)]
    color: [f32; 3],
}

pub struct Scene {
    pipeline: Arc<GraphicsPipeline>,
    vertex_buffer: Subbuffer<[MyVertex]>,
}

impl Scene {
    pub fn initialize(device: &Arc<Device>, swapchain: &Arc<Swapchain>, memory_allocator: &Arc<StandardMemoryAllocator>) -> Scene {
        mod vs {
            vulkano_shaders::shader! {
                ty: "vertex",
                path: r"res/shaders/vertex.vert",
            }
        }

        mod fs {
            vulkano_shaders::shader! {
                ty: "fragment",
                path: r"res/shaders/vertex.frag",
            }
        }

        let pipeline = {
            let vs = vs::load(device.clone())
                .unwrap()
                .entry_point("main")
                .unwrap();
            let fs = fs::load(device.clone())
                .unwrap()
                .entry_point("main")
                .unwrap();

            let vertex_input_state = MyVertex::per_vertex().definition(&vs).unwrap();

            let stages = [
                PipelineShaderStageCreateInfo::new(vs),
                PipelineShaderStageCreateInfo::new(fs),
            ];

            let layout = PipelineLayout::new(
                device.clone(),
                PipelineDescriptorSetLayoutCreateInfo::from_stages(&stages)
                    .into_pipeline_layout_create_info(device.clone())
                    .unwrap(),
            )
                .unwrap();

            let subpass = PipelineRenderingCreateInfo {
                color_attachment_formats: vec![Some(swapchain.image_format())],
                ..Default::default()
            };

            GraphicsPipeline::new(
                device.clone(),
                None,
                GraphicsPipelineCreateInfo {
                    stages: stages.into_iter().collect(),
                    vertex_input_state: Some(vertex_input_state),
                    input_assembly_state: Some(InputAssemblyState::default()),
                    viewport_state: Some(ViewportState::default()),
                    rasterization_state: Some(RasterizationState::default()),
                    multisample_state: Some(MultisampleState::default()),
                    color_blend_state: Some(ColorBlendState::with_attachment_states(
                        subpass.color_attachment_formats.len() as u32,
                        ColorBlendAttachmentState::default(),
                    )),
                    dynamic_state: [DynamicState::Viewport].into_iter().collect(),
                    subpass: Some(subpass.into()),
                    ..GraphicsPipelineCreateInfo::layout(layout)
                },
            )
                .unwrap()
        };

        let vertices = [
            // Triangle en haut à gauche
            MyVertex {
                position: [-0.5, -0.75],  // Coin supérieur gauche
                color: [1.0, 0.0, 0.0],  // Couleur rouge
            },
            MyVertex {
                position: [-0.75, -0.25],  // Coin supérieur
                color: [0.0, 1.0, 0.0],  // Couleur verte
            },
            MyVertex {
                position: [-0.25, -0.25],  // Coin supérieur droit
                color: [0.0, 0.0, 1.0],  // Couleur bleue
            },

            // Triangle en bas à gauche
            MyVertex {
                position: [-0.5, 0.25],  // Coin inférieur gauche
                color: [0.5, 0.5, 0.5],  // Couleur gris
            },
            MyVertex {
                position: [-0.75, 0.75],  // Coin inférieur
                color: [0.2, 0.8, 0.2],  // Couleur vert clair
            },
            MyVertex {
                position: [-0.25, 0.75],  // Coin inférieur droit
                color: [0.8, 0.2, 0.2],  // Couleur rouge clair
            },

            // Triangle en haut à droite
            MyVertex {
                position: [0.5, -0.75],  // Coin gauche supérieur
                color: [1.0, 1.0, 0.0],  // Couleur jaune
            },
            MyVertex {
                position: [0.25, -0.25],  // Coin gauche
                color: [0.0, 1.0, 1.0],  // Couleur cyan
            },
            MyVertex {
                position: [0.75, -0.25],  // Coin gauche inférieur
                color: [1.0, 0.0, 1.0],  // Couleur magenta
            },

            // Triangle en bas à droite
            MyVertex {
                position: [0.5, 0.25],  // Coin droit supérieur
                color: [0.1, 0.5, 0.8],  // Couleur bleu clair
            },
            MyVertex {
                position: [0.25, 0.75],  // Coin droit
                color: [0.8, 0.6, 0.1],  // Couleur or
            },
            MyVertex {
                position: [0.75, 0.75],  // Coin droit inférieur
                color: [0.3, 0.4, 0.6],  // Couleur bleu foncé
            },
        ];
        let vertex_buffer = Buffer::from_iter(
            memory_allocator.clone(),
            BufferCreateInfo {
                usage: BufferUsage::VERTEX_BUFFER,
                ..Default::default()
            },
            AllocationCreateInfo {
                memory_type_filter: MemoryTypeFilter::PREFER_DEVICE
                    | MemoryTypeFilter::HOST_SEQUENTIAL_WRITE,
                ..Default::default()
            },
            vertices,
        )
            .unwrap();

        Self {
            pipeline,
            vertex_buffer,
        }
    }

    pub fn render(&self, builder: &mut AutoCommandBufferBuilder<PrimaryAutoCommandBuffer>) {
        builder.bind_pipeline_graphics(self.pipeline.clone())
            .unwrap()
            .bind_vertex_buffers(0, self.vertex_buffer.clone())
            .unwrap();

        let vertex_count = self.vertex_buffer.len() as u32;
        let instance_count = vertex_count / 3;

        unsafe { builder.draw(vertex_count, instance_count, 0, 0) }.unwrap();
    }
}